# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Base classes for collections.
=============================
"""
from __future__ import annotations

from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)
import dataclasses
import itertools
import pathlib

import dask.bag.core
import dask.distributed
import dask.utils
import fsspec
import zarr

from .. import (
    dask_utils,
    dataset,
    expression,
    fs_utils,
    meta,
    partitioning,
    storage,
    sync,
)
from .callable_objects import MapCallable, PartitionCallable
from .detail import (
    PartitioningProperties,
    _load_and_apply_indexer,
    _load_dataset,
    _load_dataset_with_overlap,
)

#: Type of functions filtering the partitions.
PartitionFilterCallback = Callable[[Dict[str, int]], bool]

#: Type of argument to filter the partitions.
PartitionFilter = Optional[Union[str, PartitionFilterCallback]]

#: Indexer's type.
Indexer = Iterable[Tuple[Tuple[Tuple[str, int], ...], slice]]

#: Indexer arguments' type.
IndexerArgs = Tuple[Tuple[Tuple[str, int], ...], List[slice]]

#: Name of the directory storing the immutable dataset.
_IMMUTABLE = '.immutable'


def build_indexer_args(
    collection: ReadOnlyCollection,
    filters: PartitionFilter,
    indexer: Indexer,
    *,
    partitions: Iterable[str] | None = None,
) -> Iterator[IndexerArgs]:
    """Build the arguments for the indexer.

    Args:
        collection: The collection to index.
        filters: The partition filters.
        indexer: The indexer.
        partitions: The partitions to index. If None, all the partitions
            are indexed.

    Returns:
        An iterator containing the arguments for the indexer.
    """
    # Build an indexer dictionary between the partition scheme and
    # indexer.
    indexers_map: dict[tuple[tuple[str, int], ...], list[slice]] = {}
    _ = {
        indexers_map.setdefault(  # type: ignore[func-returns-value]
            partition_scheme, []).append(indexer)
        for partition_scheme, indexer in indexer
    }
    # Filter the selected partitions
    partitions = partitions or collection.partitions(filters=filters)
    selected_partitions = set(indexers_map) & {
        collection.partitioning.parse(item)
        for item in partitions
    }

    # For each provided partition scheme, retrieves the corresponding
    # indexer.
    return ((item, indexers_map[item]) for item in sorted(selected_partitions))


def _immutable_path(
    zds: meta.Dataset,
    partition_properties: PartitioningProperties,
) -> str | None:
    """Return the immutable path of the dataset.

    Args:
        zds: The dataset to process.
        partition_properties: The partitioning properties.

    Returns:
        The immutable path of the dataset containing data that are immutable
        relative to the partitioning or None if the dataset does not contain
        immutable data.
    """
    return fs_utils.join_path(
        partition_properties.dir, _IMMUTABLE) if zds.select_variables_by_dims(
            (partition_properties.dim, ), predicate=False) else None


@dataclasses.dataclass(frozen=True)
class CollectionProperties:
    """This class contains the properties of a collection."""
    #: The axis of the collection.
    axis: str

    #: The metadata that describes the dataset handled by the collection.
    metadata: meta.Dataset

    #: The partitioning strategy used to split the data.
    partition_strategy: partitioning.Partitioning

    #: The partitioning properties (base directory and dimension).
    partition: PartitioningProperties


@dataclasses.dataclass(frozen=True)
class CollectionSettings:
    """This class contains the settings of a collection."""
    #: The mode of access of the collection.
    mode: Literal['r', 'w']

    #: The file system used to read/write the collection.
    filesystem: fsspec.AbstractFileSystem

    #: The synchronizer used to synchronize the modifications.
    synchronizer: sync.Sync

    def __post_init__(self) -> None:
        if self.mode not in ('r', 'w'):
            raise ValueError(f'The mode {self.mode!r} is not supported.')


class ReadOnlyCollection:
    """Collection base class, offering read-only access to a collection of
    datasets.

    The arguments of the constructor are detailed in the documentation
    of the parent class :py:class:`zarr_collection.Collection`.
    """
    #: Configuration filename of the collection.
    CONFIG: ClassVar[str] = '.zcollection'

    def __init__(
        self,
        axis: str,
        ds: meta.Dataset,
        partition_handler: partitioning.Partitioning,
        partition_base_dir: str,
        *,
        mode: Literal['r', 'w'] | None = None,
        filesystem: fsspec.AbstractFileSystem | str | None = None,
        synchronizer: sync.Sync | None = None,
    ) -> None:
        if axis not in ds.variables:
            raise ValueError(
                f'The variable {axis!r} is not defined in the dataset.')

        for varname in partition_handler.variables:
            if varname not in ds.variables:
                raise ValueError(
                    f'The partitioning key {varname!r} is not defined in '
                    'the dataset.')

        self._settings = CollectionSettings(mode or 'w',
                                            fs_utils.get_fs(filesystem),
                                            synchronizer or sync.NoSync())

        self._properties = CollectionProperties(
            axis, ds, partition_handler,
            PartitioningProperties(
                fs_utils.normalize_path(self._settings.filesystem,
                                        partition_base_dir),
                ds.variables[axis].dimensions[0]))

        #: The path to the dataset that contains the immutable data relative
        #: to the partitioning.
        self._immutable: str | None = _immutable_path(
            ds, self._properties.partition)

    @property
    def axis(self) -> str:
        """Return the axis of the collection."""
        return self._properties.axis

    @property
    def metadata(self) -> meta.Dataset:
        """Return the metadata of the collection."""
        return self._properties.metadata

    @property
    def partitioning(self) -> partitioning.Partitioning:
        """Return the partitioning strategy of the collection."""
        return self._properties.partition_strategy

    @property
    def partition_properties(self) -> PartitioningProperties:
        """Return the partitioning properties of the collection."""
        return self._properties.partition

    @property
    def mode(self) -> Literal['r', 'w']:
        """Return the mode of the collection."""
        return self._settings.mode

    @property
    def fs(self) -> fsspec.AbstractFileSystem:
        """Return the filesystem of the collection."""
        return self._settings.filesystem

    @property
    def synchronizer(self) -> sync.Sync:
        """Return the synchronizer of the collection."""
        return self._settings.synchronizer

    @property
    def immutable(self) -> bool:
        """Return True if the collection contains immutable data relative to
        the partitioning."""
        return self._immutable is not None

    @classmethod
    def _config(cls, partition_base_dir: str) -> str:
        """Return the configuration path."""
        return fs_utils.join_path(partition_base_dir, cls.CONFIG)

    def is_locked(self) -> bool:
        """Return True if the collection is locked."""
        return self.synchronizer.is_locked()

    def _is_selected(
        self,
        partition: Sequence[str],
        expr: Callable[[dict[str, int]], bool] | None,
    ) -> bool:
        """Return whether the partition is selected.

        Args:
            partition: The partition to check.
            expr: The expression used to filter the partition.

        Returns:
            Whether the partition is selected.
        """
        return True if expr is None else expr(
            dict(self.partitioning.parse('/'.join(partition))))

    def _relative_path(self, path: str) -> str:
        """Return the relative path to the collection.

        Args:
            path: The path to the dataset.

        Returns:
            The relative path to the collection.
        """
        return pathlib.Path(path).relative_to(
            self.partition_properties.dir).as_posix()

    def partitions(
        self,
        *,
        lock: bool = False,
        filters: PartitionFilter = None,
        relative: bool = False,
    ) -> Iterator[str]:
        """List the partitions of the collection.

        Args:
            lock: Whether to lock the collection or not to avoid listing
                partitions while the collection is being modified.
            filters: The predicate used to filter the partitions to load. If
                the predicate is a string, it is a valid python expression to
                filter the partitions, using the partitioning scheme as
                variables. If the predicate is a function, it is a function that
                takes the partition scheme as input and returns a boolean.
            relative: Whether to return the relative path.

        Returns:
            The list of partitions.

        Example:
            >>> tuple(collection.partitions(
            ...     filters="year == 2019 and month == 1"))
            ('year=2019/month=01/day=01', 'year=2019/month=01/day=02/', ...)
            >>> tuple(collection.partitions(
            ...     filters=lambda x: x["year"] == 2019 and x["month"] == 1))
            ('year=2019/month=01/day=01', 'year=2019/month=01/day=02/', ...)
        """
        if isinstance(filters, str):
            expr: Any = expression.Expression(filters)
        elif callable(filters):
            expr = filters
        else:
            expr = None

        base_dir: str = self.partition_properties.dir
        sep: str = self.fs.sep

        if lock:
            with self.synchronizer:
                partitions: Iterable[str] = tuple(
                    self.partitioning.list_partitions(self.fs, base_dir))
        else:
            partitions = self.partitioning.list_partitions(self.fs, base_dir)

        yield from (self._relative_path(item) if relative else item
                    for item in partitions
                    if (item != self._immutable and self._is_selected(
                        item.replace(base_dir, '').split(sep), expr)))

    # pylint: disable=duplicate-code
    # false positive, no code duplication
    def map(
        self,
        func: MapCallable,
        /,
        *args,
        filters: PartitionFilter = None,
        partition_size: int | None = None,
        npartitions: int | None = None,
        selected_variables: Sequence[str] | None = None,
        delayed: bool = True,
        **kwargs,
    ) -> dask.bag.core.Bag:
        """Map a function over the partitions of the collection.

        Args:
            func: The function to apply to every partition of the collection.
            *args: The positional arguments to pass to the function.
            filters: The predicate used to filter the partitions to process.
                To get more information on the predicate, see the
                documentation of the :meth:`partitions` method.
            partition_size: The length of each bag partition.
            npartitions: The number of desired bag partitions.
            selected_variables: A list of variables to retain from the
                collection. If None, all variables are kept.
            delayed: Whether to load the data lazily or not.
            **kwargs: The keyword arguments to pass to the function.

        Returns:
            A bag containing the tuple of the partition scheme and the result
            of the function.

        Example:
            >>> futures = collection.map(
            ...     lambda x: (x["var1"] + x["var2"]).values)
            >>> for item in futures:
            ...     print(item)
            [1.0, 2.0, 3.0, 4.0]
            [5.0, 6.0, 7.0, 8.0]
        """

        def _wrap(
            partition: str,
            func: PartitionCallable,
            selected_variables: Sequence[str] | None,
            delayed: bool,
            *args,
            **kwargs,
        ) -> tuple[tuple[tuple[str, int], ...], Any]:
            """Wraps the function to apply on the partition.

            Args:
                func: The function to apply.
                partition: The partition to apply the function on.
                selected_variables: The list of variables to retain from the
                    partition.
                *args: The positional arguments to pass to the function.
                **kwargs: The keyword arguments to pass to the function.

            Returns:
                The result of the function.
            """
            zds: dataset.Dataset = _load_dataset(delayed, self.fs,
                                                 self._immutable, partition,
                                                 selected_variables)
            return self.partitioning.parse(partition), func(
                zds, *args, **kwargs)

        if not callable(func):
            raise TypeError('func must be a callable')

        bag: dask.bag.core.Bag = dask.bag.core.from_sequence(
            self.partitions(filters=filters),
            partition_size=partition_size,
            npartitions=npartitions)
        return bag.map(_wrap, func, selected_variables, delayed, *args,
                       **kwargs)
        # pylint: enable=duplicate-code

    def map_overlap(
        self,
        func: MapCallable,
        /,
        *args,
        delayed: bool = True,
        depth: int = 1,
        filters: PartitionFilter = None,
        partition_size: int | None = None,
        npartition: int | None = None,
        selected_variables: Sequence[str] | None = None,
        **kwargs,
    ) -> dask.bag.core.Bag:
        """Map a function over the partitions of the collection with some
        overlap.

        Args:
            func: The function to apply to every partition of the collection.
                If ``func`` accepts a partition_info as a keyword
                argument, it will be passed a tuple with the name of the
                partitioned dimension and the slice allowing getting in the
                dataset the selected partition without the overlap.
            *args: The positional arguments to pass to the function.
            delayed: Whether to load the data lazily or not.
            depth: The depth of the overlap between the partitions. Defaults
                to 1.
            filters: The predicate used to filter the partitions to process.
                To get more information on the predicate, see the
                documentation of the :meth:`partitions` method.
            partition_size: The length of each bag partition.
            npartition: The number of desired bag partitions.
            selected_variables: A list of variables to retain from the
                collection. If None, all variables are kept.
            **kwargs: The keyword arguments to pass to the function.

        Returns:
            A bag containing the tuple of the partition scheme and the result
            of the function.

        Example:
            >>> futures = collection.map_overlap(
            ...     lambda x: (x["var1"] + x["var2"]).values,
            ...     depth=1)
            >>> for item in futures:
            ...     print(item)
            [1.0, 2.0, 3.0, 4.0]
            [5.0, 6.0, 7.0, 8.0]
        """
        if not callable(func):
            raise TypeError('func must be a callable')

        add_partition_info: bool = dask.utils.has_keyword(
            func, 'partition_info')

        def _wrap(
            partition: str,
            *args,
            delayed: bool,
            depth: int,
            partitions: tuple[str, ...],
            selected_variables: Sequence[str] | None,
            wrapped_func: PartitionCallable,
            **kwargs,
        ) -> tuple[tuple[tuple[str, int], ...], Any]:
            """Wraps the function to apply on the partition.

            Args:
                *args: The positional arguments to pass to the function.
                delayed: Whether to load the data lazily or not.
                depth: The depth of the overlap between the partitions.
                partition: The partition to apply the function on.
                partitions: The partitions to apply the function on.
                selected_variables: The list of variables to retain from the
                    partition.
                wrapped_func: The function to apply.
                **kwargs: The keyword arguments to pass to the function.

            Returns:
                The result of the function.
            """
            zds: dataset.Dataset
            indices: slice

            zds, indices = _load_dataset_with_overlap(
                delayed=delayed,
                depth=depth,
                dim=self.partition_properties.dim,
                fs=self.fs,
                immutable=self._immutable,
                partition=partition,
                partitions=partitions,
                selected_variables=selected_variables)

            if add_partition_info:
                kwargs = kwargs.copy()
                kwargs['partition_info'] = (self.partition_properties.dim,
                                            indices)

            # Finally, apply the function.
            return (self.partitioning.parse(partition),
                    wrapped_func(zds, *args, **kwargs))

        partitions = tuple(self.partitions(filters=filters))
        bag: dask.bag.core.Bag = dask.bag.core.from_sequence(
            partitions, partition_size=partition_size, npartitions=npartition)
        return bag.map(_wrap,
                       *args,
                       delayed=delayed,
                       depth=depth,
                       partitions=partitions,
                       selected_variables=selected_variables,
                       wrapped_func=func,
                       **kwargs)

    def load(
        self,
        *,
        delayed: bool = True,
        filters: PartitionFilter = None,
        indexer: Indexer | None = None,
        selected_variables: Iterable[str] | None = None,
    ) -> dataset.Dataset | None:
        """Load the selected partitions.

        Args:
            delayed: Whether to load data in a dask array or not.
            filters: The predicate used to filter the partitions to load. To
                get more information on the predicate, see the documentation of
                the :meth:`partitions` method.
            indexer: The indexer to apply.
            selected_variables: A list of variables to retain from the
                collection. If None, all variables are kept.

        Returns:
            The dataset containing the selected partitions, or None if no
            partitions were selected.

        Warning:
            If you select variables to load from the collection, do not insert
            the returned dataset otherwise all skipped variables will be reset
            with fill values.

        Example:
            >>> collection = ...
            >>> collection.load(
            ...     filters="year == 2019 and month == 3 and day % 2 == 0")
            >>> collection.load(
            ...     filters=lambda keys: keys["year"] == 2019 and
            ...     keys["month"] == 3 and keys["day"] % 2 == 0)
        """
        client: dask.distributed.Client = dask_utils.get_client()
        arrays: list[dataset.Dataset]
        if indexer is None:
            selected_partitions = tuple(self.partitions(filters=filters))
            if len(selected_partitions) == 0:
                return None

            # No indexer, so the dataset is loaded directly for each
            # selected partition.
            bag: dask.bag.core.Bag = dask.bag.core.from_sequence(
                self.partitions(filters=filters),
                npartitions=dask_utils.dask_workers(client, cores_only=True))
            arrays = bag.map(storage.open_zarr_group,
                             delayed=delayed,
                             fs=self.fs,
                             selected_variables=selected_variables).compute()
        else:
            # Build the indexer arguments.
            args = tuple(build_indexer_args(self, filters, indexer))
            if len(args) == 0:
                return None

            bag = dask.bag.core.from_sequence(
                args,
                npartitions=dask_utils.dask_workers(client, cores_only=True))

            # Finally, load the selected partitions and apply the indexer.
            arrays = list(
                itertools.chain.from_iterable(
                    bag.map(
                        _load_and_apply_indexer,
                        delayed=delayed,
                        fs=self.fs,
                        partition_handler=self.partitioning,
                        partition_properties=self.partition_properties,
                        selected_variables=selected_variables,
                    ).compute()))

        array: dataset.Dataset = arrays.pop(0)
        if arrays:
            array = array.concat(arrays, self.partition_properties.dim)
        if self._immutable:
            array.merge(
                storage.open_zarr_group(self._immutable,
                                        self.fs,
                                        delayed=delayed,
                                        selected_variables=selected_variables))
        array.fill_attrs(self.metadata)
        return array

    def _bag_from_partitions(
        self,
        filters: PartitionFilter | None = None,
        **kwargs,
    ) -> dask.bag.core.Bag:
        """Return a dask bag from the partitions.

        Args:
            filters: The predicate used to filter the partitions to load.
            kwargs: The keyword arguments to pass to the method
                :meth:`partitions`.

        Returns:
            The dask bag.
        """
        partitions: list[str] = [*self.partitions(filters=filters, **kwargs)]
        return dask.bag.core.from_sequence(seq=partitions,
                                           npartitions=len(partitions))

    def iterate_on_records(
        self,
        *,
        relative: bool = False,
    ) -> Iterator[tuple[str, zarr.Group]]:
        """Iterate over the partitions and the zarr groups.

        Args:
            relative: If True, the paths are relative to the base directory.

        Returns
            The iterator over the partitions and the zarr groups.
        """
        yield from (
            (
                self._relative_path(item) if relative else item,
                zarr.open_consolidated(
                    self.fs.get_mapper(item),  # type: ignore
                    mode='r',
                )) for item in self.partitions())

    def variables(
        self,
        selected_variables: Iterable[str] | None = None
    ) -> tuple[dataset.Variable, ...]:
        """Return the variables of the collection.

        Args:
            selected_variables: The variables to return. If None, all the
                variables are returned.

        Returns:
            The variables of the collection.
        """
        return dataset.get_dataset_variable_properties(self.metadata,
                                                       selected_variables)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Handle a collection of Zarr groups.
===================================
"""
from . import merging, partitioning
from .collection import Collection
from .collection.abc import Indexer, PartitionFilter, PartitionFilterCallback
from .collection.callable_objects import (
    MapCallable,
    PartitionCallable,
    UpdateCallable,
)
from .convenience import (
    create_collection,
    create_view,
    open_collection,
    open_view,
)
from .dataset import Dataset
from .meta import Attribute
from .variable import Array, DelayedArray, Variable
from .version import __version__
from .view import View, ViewReference, ViewUpdateCallable

__all__ = (
    '__version__',
    'Array',
    'Attribute',
    'Collection',
    'create_collection',
    'create_view',
    'Dataset',
    'DelayedArray',
    'Indexer',
    'MapCallable',
    'merging',
    'open_collection',
    'open_view',
    'PartitionCallable',
    'PartitionFilter',
    'PartitionFilterCallback',
    'partitioning',
    'UpdateCallable',
    'Variable',
    'version',
    'View',
    'ViewReference',
    'ViewUpdateCallable',
)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Abstract base class for indexing.
=================================
"""
from __future__ import annotations

from typing import Any, Dict, Iterable, Protocol, Union
import abc
import functools
import pathlib

import fsspec
import numpy
import pyarrow
import pyarrow.parquet

from .. import collection, dataset
from ..collection.callable_objects import MapCallable
from ..type_hints import NDArray

#: Scalar data type for the index.
Scalar = Union[int, float, bytes]

#: Index data type.
DType = Union[Scalar, Iterable[Scalar]]

#: Type of associative dictionary used for index queries, which matches a
#: column of the index to the requested values.
QueryDict = Dict[str, DType]


#: pylint: disable=too-few-public-methods
class IndexingCallable(Protocol):
    """Protocol for indexing the partitions of a collection.

    A partition callable is a function that accepts a dataset and
    returns a numpy structured array to be converted to a DataFrame and
    stored in the index. The function is called for each partition of
    the collection to determine the first and last index of the
    partition that contains the value to be indexed.
    """

    def __call__(
        self,
        zds: dataset.Dataset,
        *args,
        **kwargs,
    ) -> NDArray:
        """Indexing the partition of the collection.

        Args:
            zds: Dataset to be indexed.
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            A numpy structured array to be converted to a DataFrame and stored
            in the index.
        """
        # pylint: disable=unnecessary-ellipsis
        # Ellipsis is necessary to make the function signature match the
        # protocol.
        ...  # pragma: no cover
        # pylint: enable=unnecessary-ellipsis


class Indexer(abc.ABC):
    """Abstract base class for indexing a collection.

    This class defines the interface for indexing a collection.

    Args:
        path: The path to the index.
        filesystem: The filesystem to use.
    """
    #: The name of the column containing the start of the slice.
    START = 'start'

    #: The name of the column containing the stop of the slice.
    STOP = 'stop'

    def __init__(
        self,
        path: pathlib.Path | str,
        *,
        filesystem: fsspec.AbstractFileSystem | None = None,
    ) -> None:
        if isinstance(path, pathlib.Path):
            path = str(path)
        #: Path to the index.
        self._path: str = path
        #: Filesystem to use.
        self._fs: fsspec.AbstractFileSystem = \
            filesystem or fsspec.filesystem('file')
        #: Metadata to attach to the index.
        self._meta: dict[str, bytes] = {}
        #: Partitioning keys of the indexed collection.
        self._partition_keys: tuple[str, ...] = ()
        #: PyArrow table containing the index.
        self._table: pyarrow.Table | None = None
        #: Type for each columns of the index.
        self._type: dict[str, pyarrow.DataType] = {}

    @property
    def meta(self) -> dict[str, bytes]:
        """Metadata attached to the index.

        Returns:
            The metadata.
        """
        return self._meta

    @classmethod
    def dtype(cls, **_kwargs) -> list[tuple[str, str]]:
        """Return the columns of the index.

        Args:
            **kwargs: Additional arguments to pass to the function.

        Returns:
            A tuple of (name, type) pairs.
        """
        return [
            (cls.START, 'int64'),
            (cls.STOP, 'int64'),
        ]

    @classmethod
    def pyarrow_type(cls, **kwargs) -> dict[str, pyarrow.DataType]:
        """Return the PyArrow DataType for the index.

        Args:
            **kwargs: Additional arguments to pass to the function.

        Returns:
            The PyArrow type.
        """
        dtype = dict(cls.dtype(**kwargs))
        binary: dict[str, pyarrow.DataType] = {}
        for name, value in tuple(dtype.items()):
            if value.startswith('S'):
                binary[name] = pyarrow.binary(int(value[1:]))
                del dtype[name]
        result: dict[str, pyarrow.DataType] = {
            name: getattr(pyarrow, value)()
            for name, value in dtype.items()
        }
        result.update(binary)
        return result

    def _set_schema(
        self,
        partition_schema: tuple[tuple[str, pyarrow.DataType], ...],
        **kwargs,
    ) -> None:
        """Set the schema properties of the index.

        Args:
            partition_schema: A tuple of (name, type) pairs that describes the
                storage properties of the collection's partitioning keys.
        """
        dtype: dict[str, pyarrow.DataType] = self.pyarrow_type(**kwargs)
        self._partition_keys = tuple(item[0] for item in partition_schema)
        self._type = {name: dtype[name] for name, _ in self.dtype()}
        self._type.update({item[0]: item[1] for item in partition_schema})

    def _sort_keys(self) -> list[tuple[str, str]]:
        """Return the list of keys to sort the index by."""
        keys: tuple[str, ...] = self._partition_keys + (self.START, self.STOP)
        return [(key, 'ascending') for key in keys]

    @classmethod
    def _create(
        cls,
        path: pathlib.Path | str,
        zds: collection.Collection,
        meta: dict[str, bytes] | None = None,
        filesystem: fsspec.AbstractFileSystem | None = None,
        **kwargs,
    ) -> Indexer:
        """Create a new index.

        Args:
            path: The path to the index.
            zds: The collection to index.
            meta: Metadata to attach to the index.
            filesystem: The filesystem to use.

        Returns:
            The created index.
        """
        partition_schema = tuple((name, getattr(pyarrow, value)())
                                 for name, value in zds.partitioning.dtype())
        self: Indexer = cls(path, filesystem=filesystem)
        self._meta = meta or {}
        self._set_schema(partition_schema, **kwargs)
        return self

    @classmethod
    @abc.abstractmethod
    def create(
        cls,
        path: pathlib.Path | str,
        zds: collection.Collection,
        *,
        filesystem: fsspec.AbstractFileSystem | None = None,
        **kwargs,
    ) -> Indexer:
        """Create a new index.

        Args:
            path: The path to the index.
            zds: The collection to index.
            filesystem: The filesystem to use.

        Returns:
            The created index.
        """

    @classmethod
    def open(
        cls,
        path: pathlib.Path | str,
        *,
        filesystem: fsspec.AbstractFileSystem | None = None,
    ) -> Indexer:
        """Open an index.

        Args:
            path: The path to the index.
            filesystem: The filesystem to use.

        Returns:
            The index.
        """
        self: Indexer = cls(path, filesystem=filesystem)
        with self._fs.open(path, 'rb') as stream:
            schema: pyarrow.Schema = pyarrow.parquet.read_schema(stream)
        columns = tuple(name for name, _ in self.dtype())
        self._partition_keys = tuple(name for name in schema.names
                                     if name not in columns)
        self._type = {name: schema.field(name).type for name in schema.names}
        self._meta = {
            name.decode(): value
            for name, value in schema.metadata.items()
        } if schema.metadata is not None else {}
        return self

    def _update(
        self,
        zcollection: collection.Collection,
        func: MapCallable,
        partition_size: int | None = None,
        npartitions: int | None = None,
        **kwargs,
    ) -> None:
        """Update the index.

        Args:
            zcollection: The collection to index.
            func: The function to use to calculate the index.
            partition_size: The length of each bag partition.
            npartitions: The number of desired bag partitions.
            **kwargs: Additional arguments to pass to the function.
        """
        tables: list[pyarrow.Table] = []
        bag = zcollection.map(func,
                              partition_size=partition_size,
                              npartitions=npartitions,
                              **kwargs)
        # List of new partitions indexed.
        partitions = []
        for partition, data in bag.compute():
            length = data.size
            # If the current item is empty, skip it.
            if length == 0:
                continue
            # Create a new table with the indexed data.
            data = {
                field: pyarrow.array(data[field], type=self._type[field])
                for field in data.dtype.fields
            }
            # Add the partition to the table.
            data.update(
                (name,
                 pyarrow.nulls(length, type=self._type[name]).fill_null(value))
                for name, value in partition)
            # Memoize the updated partitions.
            partitions.append(tuple(
                (name, value) for name, value in partition))
            tables.append(pyarrow.Table.from_pydict(data))

        # If no new data, skip the update.
        if not partitions:
            return

        # The existing index must be updated?
        if self._fs.exists(self._path):
            table = pyarrow.parquet.read_table(self._path, filesystem=self._fs)

            # pylint: disable=no-member
            # Build the list of mask to select the rows to drop.
            mask = []
            for item in partitions:
                mask.append(
                    functools.reduce(
                        pyarrow.compute.and_,  # type:ignore
                        [
                            pyarrow.compute.is_in(  # type:ignore
                                table[name],
                                value_set=pyarrow.array([value],
                                                        type=self._type[name]))
                            for name, value in item
                        ]))
            mask = functools.reduce(pyarrow.compute.or_, mask)  # type:ignore

            # Inserts the previous index without the updated partitions.
            tables.insert(0, table.filter(
                pyarrow.compute.invert(mask)))  # type:ignore
            # pylint: enable=no-member

        if len(tables) == 0:
            # No new data, nothing to do.
            return

        table = pyarrow.concat_tables(tables)
        if len(self._meta) and table.schema.metadata is None:
            table = table.replace_schema_metadata(self._meta)
        pyarrow.parquet.write_table(table, self._path, filesystem=self._fs)
        self._table = table

    @abc.abstractmethod
    def update(
        self,
        zds: collection.Collection,
        *,
        partition_size: int | None = None,
        npartitions: int | None = None,
    ) -> None:
        """Update the index.

        Args:
            zds: The dataset containing the new data.
            partition_size: The length of each bag partition.
            npartitions: The number of desired bag partitions.
        """

    def _read(self) -> pyarrow.Table:
        """Read the index."""
        if self._table is None:
            self._table = pyarrow.parquet.read_table(self._path,
                                                     filesystem=self._fs)
        return self._table

    def _table_2_indexer(self, table: pyarrow.Table,
                         only_partition_keys: bool) -> collection.abc.Indexer:
        """Convert a table to an indexer.

        Args:
            table: The table to convert.
            only_partition_keys: If True, only the partition keys are kept.

        Returns:
            The indexer.
        """
        # Columns to keep. START and STOP are always dropped, as they're
        # exported within the slice.
        column_names = (self._partition_keys if only_partition_keys else tuple(
            filter(lambda item: item not in [self.START, self.STOP],
                   self.table.column_names)))

        # Convert columns of partitioning to numpy arrays.
        data = {name: table[name].to_numpy() for name in column_names}

        # Convert columns of the slice definition to numpy arrays.
        start = table[self.START].to_numpy()
        stop = table[self.STOP].to_numpy()

        # Calculate the indexes of each contiguous slice.
        # index = [0,     1, ...,   69,    70, ...,   134]
        # start = [817, 823, ..., 2320, 17333, ..., 19337]
        # stop  = [832, 874, ..., 2396, 17420, ..., 19389]
        # chunks = [0, 70, 134]
        chunks: Any = [[0],
                       numpy.where(start[1:] - stop[:-1] > 0)[0] + 1,
                       [len(table)]]

        # Adds the chunks corresponding to the partitioning keys
        chunks += [
            numpy.where(numpy.roll(values, 1) != values)[0]
            for values in data.values()
        ]

        # Finally, we build the list of indexes of the different chunks found.
        chunks = numpy.unique(numpy.concatenate(chunks))

        return (
            tuple(  # type:ignore
                (tuple(
                    (name, data[name][ix0])
                    for name in column_names), slice(start[ix0],
                                                     stop[ix1 - 1])), )
            for ix0, ix1 in tuple(zip(chunks[:-1], chunks[1:])))

    def query(
        self,
        columns: QueryDict,
        *,
        logical_op: str | None = None,
        mask: pyarrow.ChunkedArray | None = None,
        only_partition_keys: bool = True,
    ) -> collection.abc.Indexer:
        """Query the index.

        Args:
            columns: Dictionary of columns to query.
            logical_op: The logical operator to use. Can be "and", "and_not",
                "invert", "or", "xor". Defaults to "and".
            mask: An optional mask to apply to the table before querying.
            only_partition_keys: If True, only the partition keys are kept.

        Returns:
            Indexer.
        """
        if len(self._partition_keys) == 0:
            return tuple()

        logical_op = logical_op or 'and'
        if logical_op not in ('and', 'and_not', 'invert', 'or', 'xor'):
            raise ValueError(f'Invalid logical operator: {logical_op}')
        if logical_op in ('and', 'or'):
            logical_op += '_'
        function = getattr(pyarrow.compute, logical_op)

        if columns and (not set(self._type) & set(columns.keys())):
            raise ValueError(
                f'Invalid column names: {", ".join(columns.keys())}')

        # Transform the columns values into a list if they are not iterable.
        values = {
            k: [v] if not isinstance(v, Iterable) else v
            for k, v in columns.items()
        }

        table = self._read()

        if mask is not None:
            table = table.filter(mask)

        # pylint: disable=no-member
        if values:
            mask = functools.reduce(function, [
                pyarrow.compute.is_in(table[name],
                                      value_set=pyarrow.array(
                                          value, type=self._type[name]))
                for name, value in values.items()
            ])
            # pylint: disable=no-member
            table = table.filter(mask)

        # The selected table is sorted by the partitioning keys and the slice.
        table = pyarrow.compute.take(
            table,
            pyarrow.compute.sort_indices(table, sort_keys=self._sort_keys()))
        return self._table_2_indexer(table, only_partition_keys)

    @property
    def table(self) -> pyarrow.Table:
        """The index table.

        Returns:
            The index table.

        Raises:
            ValueError: If the index is not initialized.
        """
        if self._fs.exists(self._path):
            return self._read()
        raise ValueError('The index is not initialized.')



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Representation of dataset objects.
==================================
"""
from __future__ import annotations

from typing import Any, Iterable, Iterator, Sequence

from .meta import Attribute


def dimensions(dims: dict[str, int]) -> str:
    """Returns a string representation of the dimensions.

    Args:
        dims: A dictionary containing the dimensions.

    Returns:
        A string representation of the dimensions in the form of a tuple, where
        each element of the tuple is a string containing the dimension name and
        its corresponding value.
    """
    return str(tuple(f'{name}: {value}' for name, value in dims.items()))


def _maybe_truncate(obj: Any, max_size: int) -> str:
    """Truncate the string representation of an object to the given length.

    Args:
        obj: An object.
        max_size: The maximum length of the string representation.

    Returns:
        The string representation of the object, truncated to the given length
        if necessary.
    """
    result = str(obj)
    if len(result) > max_size:
        return result[:max_size - 3] + '...'
    return result


def pretty_print(obj: Any, num_characters: int = 120) -> str:
    """Return a pretty printed string representation of the given object.

    Args:
        obj:
            An object to be pretty printed.
        num_characters:
            An integer representing the maximum number of
            characters per line.

    Returns:
        A string representation of the object, pretty printed with a maximum of
        `num_characters` characters per line.
    """
    result: str = _maybe_truncate(obj, num_characters)
    return result + ' ' * max(num_characters - len(result), 0)


def calculate_column_width(items: Iterable) -> int:
    """Calculate the maximum width of a column.

    Args:
        items: An iterable of items.

    Returns:
        The maximum width of a column.
    """
    max_name: int = max(len(str(name)) for name in items)
    return max(max_name, 7)


def attributes(attrs: Sequence[Attribute]) -> Iterator[str]:
    """Get the string representation of the attributes.

    Args:
        attrs: The attributes.

    Returns:
        The string representation of the attributes.
    """
    width: int = calculate_column_width(item.name for item in attrs)
    for attr in attrs:
        name_str: str = f'    {attr.name:<{width}s}'
        yield pretty_print(f'{name_str}: {attr.value!r}')



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Tests of the expression evaluation
==================================
"""
from __future__ import annotations

import timeit

import numpy
import pytest
import xarray

from .. import dataset
from ..expression import Expression
from ..partitioning import Date
# pylint: disable=unused-import # Need to import for fixtures
from .cluster import dask_client, dask_cluster

# pylint enable=unused-import


def make_dataset(num_samples: int | None = None) -> dataset.Dataset:
    """Creation of a data set for testing purposes."""
    dates = numpy.arange(numpy.datetime64('2000-01-01', 'ns'),
                         numpy.datetime64('2009-12-31', 'ns'),
                         numpy.timedelta64(1, 'h')).astype('datetime64[ns]')
    if num_samples is not None:
        dates = dates[:num_samples + 1]
    observation = numpy.random.rand(dates.size)  # type: ignore
    return dataset.Dataset.from_xarray(
        xarray.Dataset({
            'dates':
            xarray.DataArray(dates, dims=('num_lines', )),
            'observation':
            xarray.DataArray(observation, dims=('num_lines', ))
        }))


def test_expression() -> None:
    """Test of the creation of expressions."""
    expr = Expression('a == b')
    assert expr({'a': 1, 'b': 1})
    assert not expr({'a': 1, 'b': 2})

    with pytest.raises(SyntaxError):
        Expression('a==')

    with pytest.raises(NameError):
        assert expr({'a': 1, 'c': 1})


def test_date_expression(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test of expressions handling dates.."""
    zds = make_dataset(5 * 24)
    partitioning = Date(('dates', ), 'D')

    for partition, _ in partitioning.split_dataset(zds, 'num_lines'):
        variables = dict(partitioning.parse('/'.join(partition)))
        expr = Expression('year==2000')
        assert expr(variables)
        expr = Expression('year==2000 and month==1')
        assert expr(variables)
        expr = Expression('year==2000 and month==1 and day in range(1, 12)')
        assert expr(variables)


def test_bench_expression(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Benchmark of expressions."""
    partitioning = Date(('dates', ), 'D')
    zds = make_dataset()
    expr = Expression('year==2000 and month==1 and day in range(1, 12)')
    times = []
    number = 100
    for partition, _ in partitioning.split_dataset(zds, 'num_lines'):
        variables = dict(partitioning.parse('/'.join(partition)))
        times.append(
            timeit.timeit('expr(variables)',
                          globals={
                              'expr': expr,
                              'variables': variables
                          },
                          number=number))

    assert sum(times) / (len(times) * number) < 1e-5



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Delayed variable arrays.
========================
"""
from __future__ import annotations

from typing import Any, Callable, Mapping, MutableMapping, Sequence
import uuid

import dask.array.core
import dask.array.creation
import dask.array.ma
import dask.base
import dask.highlevelgraph
import dask.threaded
import numcodecs.abc
import numpy
import zarr

from ..meta import Attribute
from ..type_hints import ArrayLike, NDArray, NDMaskedArray
from .abc import Variable, concat, new_variable, not_equal

#: The dask array getter used to access the data.
GETTER: Callable = dask.array.core.getter


def _blockdims_from_blockshape(
        shape: tuple[int, ...],
        chunks: tuple[int, ...]) -> tuple[tuple[int, ...], ...]:
    """Convert a blockshape to a blockdims tuple."""
    return tuple(((chunk_item, ) * (shape_item // chunk_item) +
                  ((shape_item % chunk_item, ) if shape_item %
                   chunk_item else ()) if shape_item else (0, ))
                 for shape_item, chunk_item in zip(shape, chunks))


def from_zarr_array(
    array: zarr.Array,
    shape: Sequence[int],
    chunks: Sequence[int],
    name: str,
    *,
    lock: bool = False,
    asarray=True,
    inline_array=True,
) -> dask.array.core.Array:
    """Convert a Zarr array to a Dask array with the specified shape and
    chunks.

    Args:
        array: The Zarr array to convert.
        shape: The desired shape of the resulting Dask array.
        chunks: The desired chunk shape of the resulting Dask array.
        name: The name of the resulting Dask array.
        lock: Whether to use a lock to protect the underlying data store.
        asarray: Whether to return the resulting Dask array as an array.
        inline_array: Whether to inline the resulting array data in the tasks.

    Returns:
        A Dask array equivalent to the provided Zarr array.
    """
    dsk: dask.highlevelgraph.HighLevelGraph
    normalized_chunks: tuple[tuple[int, ...], ...]

    normalized_chunks = sum(  # type: ignore[return-value]
        (_blockdims_from_blockshape(
            (shape_item, ),
            (chunk_item, )) if not isinstance(chunk_item, (tuple, list)) else
         (chunk_item, ) for shape_item, chunk_item in zip(shape, chunks)),
        (),
    )
    dsk = dask.array.core.graph_from_arraylike(
        array,
        normalized_chunks,
        shape,
        name,
        getitem=GETTER,
        lock=lock,
        asarray=asarray,
        dtype=array.dtype,
        inline_array=inline_array,
    )
    return dask.array.core.Array(dsk,
                                 name,
                                 normalized_chunks,
                                 meta=array,
                                 dtype=array.dtype)


def _as_dask_array(
    arr: Any,
    *,
    fill_value: Any | None = None,
) -> tuple[dask.array.core.Array, Any]:
    """Convert an array-like object to a dask array.

    Args:
        arr: An array-like object.
        fill_value: The fill value.

    Returns:
        If the data provided is a masked array, the functions return the array
        with masked data replaced by its fill value and the fill value of the
        offered masked array. Otherwise, the provided array and fill value.
    """
    result: dask.array.core.Array = dask.array.core.asarray(arr)
    _meta: Any = result._meta  # pylint: disable=protected-access
    if isinstance(_meta, numpy.ma.MaskedArray):
        if fill_value is not None and not_equal(fill_value, _meta.fill_value):
            raise ValueError(
                f'The fill value {fill_value!r} does not match the fill value '
                f'{_meta.fill_value!r} of the array.')
        return dask.array.ma.filled(result, _meta.fill_value), _meta.fill_value
    return result, fill_value


class ModifiedVariableError(RuntimeError):
    """Raised when a variable has been modified since is was initialized."""

    def __str__(self) -> str:
        """Get the string representation of the exception.

        Returns:
            The string representation of the exception.
        """
        return ('You tried to access the data of a variable that has been '
                'modified since its initialization. Try to re-load the '
                'dataset.')


class DelayedArray(Variable):
    """Access to the chunked data using Dask arrays.

    Args:
        name: Name of the variable
        data: Variable data
        dimensions: Variable dimensions
        attrs: Variable attributes
        compressor: Compression codec
        fill_value: Value to use for uninitialized values
        filters: Filters to apply before writing data to disk
    """

    def __init__(self,
                 name: str,
                 data: ArrayLike[Any],
                 dimensions: Sequence[str],
                 *,
                 attrs: Sequence[Attribute] | None = None,
                 compressor: numcodecs.abc.Codec | None = None,
                 fill_value: Any | None = None,
                 filters: Sequence[numcodecs.abc.Codec] | None = None) -> None:
        array: dask.array.core.Array
        array, fill_value = _as_dask_array(data, fill_value=fill_value)
        # pylint: disable=duplicate-code
        # The code is not duplicated, we need to call the parent constructor,
        # but pylint does not understand that.
        super().__init__(
            name,
            array,
            dimensions,
            attrs=attrs,
            compressor=compressor,
            fill_value=fill_value,
            filters=filters,
        )
        # pylint: enable=duplicate-code

    @property
    def data(self) -> dask.array.core.Array:
        """Return the underlying dask array where values equal to the fill
        value are masked. If no fill value is set, the returned array is the
        same as the underlying array.

        Returns:
            The dask array

        .. seealso::

            :meth:`Variable.array`
        """
        # If the fill value is None, or if the dask array already holds a
        # masked array, return the underlying array.
        # pylint: disable=protected-access
        # No other way to check if the dask array is a masked array.
        if (self.fill_value is None
                or isinstance(self.array._meta, numpy.ma.MaskedArray)):
            return self.array
        # pylint: enable=protected-access
        return dask.array.ma.masked_equal(self.array, self.fill_value)

    @property
    def values(self) -> NDArray | NDMaskedArray:
        """Return the variable data as a numpy array.

        .. note::

            If the variable has a fill value, the result is a masked array where
            masked values are equal to the fill value.

        Returns:
            The variable data
        """
        return self.compute()

    @values.setter
    def values(self, data: ArrayLike[Any]) -> None:
        """Defines the underlying dask array. If the data provided is a masked
        array, it's converted to an array, where the masked values are replaced
        by its fill value, and its fill value becomes the new fill value of
        this instance. Otherwise, the underlying array is defined as the new
        data and the fill value is set to None.

        Args:
            data: The new data to use

        Raises:
            ValueError: If the shape of the data does not match the shape of
                the stored data.
        """
        if len(data.shape) != len(self.dimensions):
            raise ValueError('data shape does not match variable dimensions')
        self.array, self.fill_value = _as_dask_array(
            data, fill_value=self.fill_value)

    def persist(self, **kwargs) -> DelayedArray:
        """Persist the variable data into memory.

        Args:
            **kwargs: Keyword arguments passed to
                :meth:`dask.array.Array.persist`.

        Returns:
            The variable
        """
        self.array = dask.base.persist(self.array, **kwargs)
        return self

    def compute(self, **kwargs) -> NDArray | NDMaskedArray:
        """Return the variable data as a numpy array.

        .. note::

            If the variable has a fill value, the result is a masked array where
            masked values are equal to the fill value.

        Args:
            **kwargs: Keyword arguments passed to
                :meth:`dask.array.Array.compute`.
        """
        try:
            values: NDArray
            values, = dask.base.compute(self.array, traverse=False, **kwargs)
        except ValueError as exc:
            msg = str(exc)
            if 'cannot reshape' in msg or 'buffer too small' in msg:
                raise ModifiedVariableError() from exc
            raise
        return values if self.fill_value is None else numpy.ma.masked_equal(
            values, self.fill_value)

    def fill(self) -> DelayedArray:
        """Fill the variable with the fill value. If the variable has no fill
        value, this method does nothing.

        Returns:
            The variable.
        """
        if self.fill_value is not None:
            self.array = dask.array.creation.full_like(self.array,
                                                       self.fill_value)
        return self

    @classmethod
    def from_zarr(cls, array: zarr.Array, name: str, dimension: str,
                  **kwargs) -> DelayedArray:
        """Create a new variable from a zarr array.

        Args:
            array: The zarr array
            name: Name of the variable
            dimension: Name of the attribute that defines the dimensions of the
                variable
            **kwargs: Keyword arguments passed to
                :func:`dask.array.from_array`

        Returns:
            The variable
        """
        attrs = tuple(
            Attribute(k, v) for k, v in array.attrs.items() if k != dimension)
        data: dask.array.core.Array = from_zarr_array(
            array,
            array.shape,
            array.chunks,
            name=f'{name}-{uuid.uuid1()}',
            **kwargs,
        )
        # pylint: disable=duplicate-code
        # This call is similar to the one in array.py but it's not the same
        # behaviour.
        return new_variable(cls,
                            name=name,
                            array=data,
                            dimensions=array.attrs[dimension],
                            attrs=attrs,
                            compressor=array.compressor,
                            fill_value=array.fill_value,
                            filters=tuple(array.filters or ()))
        # pylint: enable=duplicate-code

    def concat(self, other: DelayedArray | Sequence[DelayedArray],
               dim: str) -> DelayedArray:
        """Concatenate this variable with another variable or a list of
        variables along a dimension.

        Args:
            other: Variable or list of variables to concatenate with this
                variable.
            dim: Dimension to concatenate along.

        Returns:
            New variable.

        Raises:
            ValueError: if the variables provided is an empty sequence.
        """
        return concat(self, other, dask.array.core.concatenate, dim)

    def __getitem__(self, key: Any) -> Any:
        """Get a slice of the variable.

        Args:
            key: Slice or index to use.
        Returns:
            The variable slice.
        """
        return self.data[key]

    def rechunk(self, **kwargs) -> DelayedArray:
        """Rechunk the variable.

        Args:
            **kwargs: Keyword arguments passed to
                :func:`dask.array.rechunk`

        Returns:
            The variable.
        """
        # pylint: disable=duplicate-code
        # False positive with the method concat.
        return new_variable(type(self),
                            name=self.name,
                            array=self.array.rechunk(**kwargs),
                            dimensions=self.dimensions,
                            attrs=self.attrs,
                            compressor=self.compressor,
                            fill_value=self.fill_value,
                            filters=self.filters)
        # pylint: enable=duplicate-code

    def to_dask_array(self):
        """Return the underlying dask array.

        Returns:
            The underlying dask array

        .. seealso::

            :func:`dask.array.asarray`
        """
        return self.data

    def __dask_graph__(self) -> Mapping | None:
        """Return the dask Graph."""
        return self.data.__dask_graph__()

    def __dask_keys__(self) -> list:
        """Return the output keys for the Dask graph."""
        return self.data.__dask_keys__()

    def __dask_layers__(self) -> tuple:
        """Return the layers for the Dask graph."""
        return self.data.__dask_layers__()

    def __dask_tokenize__(self) -> Any:
        """Return the token for the Dask graph."""
        return dask.base.normalize_token(
            (type(self), self.name, self.data, self.dimensions, self.attrs,
             self.fill_value))

    @staticmethod
    def __dask_optimize__(dsk: MutableMapping, keys: list,
                          **kwargs) -> MutableMapping:
        """Returns whether the Dask graph can be optimized.

        .. seealso::
            :meth:`dask.array.Array.__dask_optimize__`
        """
        return dask.array.core.Array.__dask_optimize__(dsk, keys, **kwargs)

    #: The default scheduler get to use for this object.
    __dask_scheduler__ = staticmethod(dask.threaded.get)

    def _dask_finalize(self, results, array_func, *args,
                       **kwargs) -> DelayedArray:
        """Finalize the computation of the variable."""
        array: Any = array_func(results, *args, **kwargs)
        if not isinstance(array, dask.array.core.Array):
            array = dask.array.core.from_array(array)
        # pylint: disable=duplicate-code
        # False positive with the method metadata defined in the base class.
        return new_variable(type(self),
                            name=self.name,
                            array=array,
                            dimensions=self.dimensions,
                            attrs=self.attrs,
                            compressor=self.compressor,
                            fill_value=self.fill_value,
                            filters=self.filters)
        # pylint: enable=duplicate-code

    def __dask_postcompute__(self) -> tuple:
        """Return the finalizer and extra arguments to convert the computed
        results into their in-memory representation."""
        array_func: Callable
        array_args: tuple

        array_func, array_args = self.data.__dask_postcompute__()
        return self._dask_finalize, (array_func, ) + array_args

    def __dask_postpersist__(self) -> tuple:
        """Return the rebuilder and extra arguments to rebuild an equivalent
        Dask collection from a persisted or rebuilt graph."""
        array_func: Callable
        array_args: tuple

        array_func, array_args = self.data.__dask_postpersist__()
        return self._dask_finalize, (array_func, ) + array_args



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Implementation details.
=======================
"""
from __future__ import annotations

from typing import Any, Callable, Dict, Iterable, Sequence, Tuple
import dataclasses
import sys
import time
import traceback
import types

import dask.utils
import fsspec
import zarr.storage

from .. import dataset, merging, partitioning, sync
from ..fs_utils import join_path
from ..storage import open_zarr_group, update_zarr_array, write_zarr_group
from ..type_hints import ArrayLike
from .callable_objects import UpdateCallable, WrappedPartitionCallable

#: Partition's type.
PartitionSlice = Tuple[Tuple[str, ...], Dict[str, slice]]


@dataclasses.dataclass(frozen=True)
class PartitioningProperties:
    """Properties of a partition."""
    #: The base directory of the partition.
    dir: str
    #: The name of the partitioning dimension.
    dim: str


def _get_slices(variable: dataset.Variable, dim: str,
                indices: slice) -> tuple[slice, ...]:
    """Return a tuple of slices that can be used to select the given dimension
    indices.

    Args:
        dim: Dimension to select
        indices: Dimension indices to select

    Returns:
        The slices
    """
    slices: list[slice] = [slice(None)] * len(variable.dimensions)
    slices[variable.dimensions.index(dim)] = indices
    return tuple(slices)


def _try_infer_callable(
    func: Callable,
    zds: dataset.Dataset,
    dim: str,
    *args,
    **kwargs,
) -> Any:
    """Try to call a function with the given arguments.

    Args:
        func: Function to call.
        zds: Dataset to pass to the function.
        dim: Name of the partitioning dimension.
        *args: Positional arguments to pass to the function.
        **kwargs: Keyword arguments to pass to the function.

    Returns:
        The result of the function.
    """
    partition_info: tuple[str, slice]
    partition_info = dim, slice(0, zds.dimensions[dim])
    try:
        if dask.utils.has_keyword(func, 'partition_info'):
            return func(zds, *args, partition_info=partition_info, **kwargs)
        return func(zds, *args, **kwargs)
    except Exception as exc:
        exc_type: type[BaseException] | None
        exc_traceback: types.TracebackType | None
        exc_value: BaseException | None
        exc_type, exc_value, exc_traceback = sys.exc_info()
        execution_tb: str = ''.join(traceback.format_tb(exc_traceback))
        raise RuntimeError(
            f'An error occurred while applying the function {func} with '
            f'the arguments {args} and {kwargs} to the partition. The '
            f'error is: {exc_type!r}: {exc_value}. The traceback '
            f'is: {execution_tb}') from exc


def _update_with_overlap(
    *args,
    func: UpdateCallable,
    zds: dataset.Dataset,
    indices: slice,
    dim: str,
    fs: fsspec.AbstractFileSystem,
    path: str,
    trim: bool,
    **kwargs,
) -> None:
    """Update a partition with overlap.

    Args:
        func: Function to apply to update each partition.
        zds: Dataset to update.
        indices: Indices of the partition to update.
        dim: Name of the partitioning dimension.
        fs: File system on which the Zarr dataset is stored.
        path: Path to the Zarr group.
        trim: Whether to trim the overlap.
        *args: Positional arguments to pass to the function.
        **kwargs: Keyword arguments to pass to the function.

    Returns:
        The updated variables.
    """
    dictionary: dict[str, ArrayLike] = (func(
        zds, *args, partition_info=(dim, indices), **
        kwargs) if dask.utils.has_keyword(func, 'partition_info') else func(
            zds, *args, **kwargs))

    if trim:
        for varname, array in dictionary.items():
            slices: tuple[slice, ...] = _get_slices(zds[varname], dim, indices)
            update_zarr_array(
                dirname=join_path(path, varname),
                array=array[slices],  # type: ignore[index]
                fs=fs,
            )
    else:
        tuple(
            map(
                lambda items: update_zarr_array(
                    dirname=join_path(path, items[0]),
                    array=items[1],
                    fs=fs,
                ), dictionary.items()))


def _load_dataset(
    delayed: bool,
    fs: fsspec.AbstractFileSystem,
    immutable: str | None,
    partition: str,
    selected_variables: Iterable[str] | None,
) -> dataset.Dataset:
    """Load a dataset from a partition.

    Args:
        delayed: Whether to load the dataset lazily.
        fs: File system on which the Zarr dataset is stored.
        immutable: Name of the immutable directory.
        partition: Name of the partition.
        selected_variables: Name of the variables to load from the dataset.

    Returns:
        The loaded dataset.
    """
    zds: dataset.Dataset = open_zarr_group(
        partition, fs, delayed=delayed, selected_variables=selected_variables)
    if immutable:
        zds.merge(
            open_zarr_group(immutable,
                            fs,
                            delayed=delayed,
                            selected_variables=selected_variables))
    return zds


def _load_dataset_with_overlap(
    *,
    delayed: bool,
    depth: int,
    dim: str,
    fs: fsspec.AbstractFileSystem,
    immutable: str | None,
    partition: str,
    partitions: Sequence[str],
    selected_variables: Iterable[str] | None,
) -> tuple[dataset.Dataset, slice]:
    """Load a dataset from a partition with overlap.

    Args:
        delayed: Whether to load the dataset lazily.
        depth: Depth of the overlap.
        dim: Name of the partitioning dimension.
        fs: File system on which the Zarr dataset is stored.
        immutable: Name of the immutable directory.
        partition: Name of the partition.
        partitions: List of all partitions.
        selected_variables: Name of the variables to load from the dataset.

    Returns:
        The loaded dataset and the slice to select the data without the
        overlap.
    """

    def calculate_slice(
        groups: list[dataset.Dataset],
        selected_partitions: list[str],
    ) -> slice:
        """Compute the slice of the selected dataset (without overlap)."""
        start = 0
        indices = slice(0, 0, None)
        for idx, zds in enumerate(groups):
            size: int = zds.dimensions[dim]
            indices = slice(start, start + size, None)
            if partition == selected_partitions[idx]:
                break
            start += size
        return indices

    where: int = partitions.index(partition)

    # Search for the overlapping partitions
    selected_partitions: list[str] = [
        partitions[ix] for ix in range(where - depth, where + depth + 1)
        if 0 <= ix < len(partitions)
    ]

    # Load the datasets for each selected partition.
    groups: list[dataset.Dataset] = [
        open_zarr_group(partition,
                        fs,
                        delayed=delayed,
                        selected_variables=selected_variables)
        for partition in selected_partitions
    ]

    # Compute the slice of the given partition.
    indices: slice = calculate_slice(groups, selected_partitions)

    # Build the dataset for the selected partitions.
    zds: dataset.Dataset = groups.pop(0)
    if groups:
        zds = zds.concat(groups, dim)

    if immutable:
        zds.merge(
            open_zarr_group(partition,
                            fs,
                            delayed=delayed,
                            selected_variables=selected_variables))
    return zds, indices


def _wrap_update_func(
    *args,
    delayed: bool,
    func: UpdateCallable,
    fs: fsspec.AbstractFileSystem,
    immutable: str | None,
    selected_variables: Iterable[str] | None,
    **kwargs,
) -> WrappedPartitionCallable:
    """Wrap an update function taking a partition's dataset as input and
    returning variable's values as a numpy array.

    Args:
        delayed: Whether to load the dataset lazily.
        func: Function to apply to update each partition.
        fs: File system on which the Zarr dataset is stored.
        immutable: Name of the immutable directory.
        selected_variables: Name of the variables to load from the dataset.
            If None, all variables are loaded.
        trim: Whether to trim the overlap.
        *args: Positional arguments to pass to the function.
        **kwargs: Keyword arguments to pass to the function.

    Returns:
        The wrapped function that takes a set of dataset partitions and the
        variable name as input and returns the variable's values as a numpy
        array.
    """

    def wrap_function(partitions: Iterable[str]) -> None:
        # Applying function for each partition's data
        for partition in partitions:
            zds: dataset.Dataset = _load_dataset(delayed, fs, immutable,
                                                 partition, selected_variables)
            dictionary: dict[str, ArrayLike] = func(zds, *args, **kwargs)
            tuple(
                update_zarr_array(  # type: ignore[func-returns-value]
                    dirname=join_path(partition, varname),
                    array=array,
                    fs=fs,
                ) for varname, array in dictionary.items())

    return wrap_function


def _wrap_update_func_with_overlap(
    *args,
    delayed: bool,
    depth: int,
    dim: str,
    func: UpdateCallable,
    fs: fsspec.AbstractFileSystem,
    immutable: str | None,
    selected_partitions: Sequence[str],
    selected_variables: Iterable[str] | None,
    trim: bool,
    **kwargs,
) -> WrappedPartitionCallable:
    """Wrap an update function taking a partition's dataset as input and
    returning variable's values as a numpy array.

    Args:
        delayed: Whether to load the dataset lazily.
        depth: Depth of the overlap.
        dim: Name of the partitioning dimension.
        func: Function to apply to update each partition.
        fs: File system on which the Zarr dataset is stored.
        immutable: Name of the immutable directory.
        selected_partitions: List of all partitions selected for the update.
        selected_variables: Name of the variables to load from the dataset.
            If None, all variables are loaded.
        trim: Whether to trim the overlap.
        *args: Positional arguments to pass to the function.
        **kwargs: Keyword arguments to pass to the function.

    Returns:
        The wrapped function that takes a set of dataset partitions and the
        variable name as input and returns the variable's values as a numpy
        array.
    """
    if depth < 0:
        raise ValueError('Depth must be non-negative.')

    def wrap_function(partitions: Sequence[str]) -> None:
        # Applying function for each partition's data
        for partition in partitions:

            # pylint: disable=duplicate-code
            # False positive with the method Collection.map_overlap
            zds: dataset.Dataset
            indices: slice
            zds, indices = _load_dataset_with_overlap(
                delayed=delayed,
                depth=depth,
                dim=dim,
                fs=fs,
                immutable=immutable,
                partition=partition,
                partitions=selected_partitions,
                selected_variables=selected_variables)
            # pylint: enable=duplicate-code

            _update_with_overlap(*args,
                                 func=func,
                                 zds=zds,
                                 indices=indices,
                                 dim=dim,
                                 fs=fs,
                                 path=partition,
                                 trim=trim,
                                 **kwargs)

    return wrap_function


def _rm(fs: fsspec.AbstractFileSystem, dirname: str) -> None:
    """Remove a directory and its content.

    Args:
        fs: The file system on which the directory is stored.
        dirname: The name of the directory to remove.
    """
    tries = 0
    while tries < 10:
        try:
            fs.rm(dirname, recursive=True)
            fs.invalidate_cache(dirname)
            if not fs.exists(dirname):
                return
        except OSError:
            fs.invalidate_cache(dirname)
        time.sleep(1)
        tries += 1


def _insert(
    *,
    args: PartitionSlice,
    axis: str,
    zds: dataset.Dataset,
    fs: fsspec.AbstractFileSystem,
    merge_callable: merging.MergeCallable | None,
    partitioning_properties: PartitioningProperties,
) -> None:
    """Insert or update a partition in the collection.

    Args:
        args: Tuple containing the partition's name and its slice.
        axis: The axis to merge on.
        zds: The dataset to process.
        fs: The file system that the partition is stored on.
        merge_callable: The merge callable.
        partitioning_properties: The partitioning properties.
    """
    partition: tuple[str, ...]
    indexer: dict[str, slice]

    partition, indexer = args
    dirname: str = join_path(*((partitioning_properties.dir, ) + partition))

    # If the consolidated zarr metadata does not exist, we consider the
    # partition as empty.
    if fs.exists(join_path(dirname, '.zmetadata')):
        # The current partition already exists, so we need to merge
        # the dataset.
        merging.perform(zds.isel(indexer),
                        dirname,
                        axis,
                        fs,
                        partitioning_properties.dim,
                        delayed=zds.delayed,
                        merge_callable=merge_callable)
        return

    # The current partition does not exist, so we need to create
    # it and insert the dataset.
    try:
        zarr.storage.init_group(store=fs.get_mapper(dirname))

        # The synchronization is done by the caller.
        write_zarr_group(zds.isel(indexer), dirname, fs, sync.NoSync())
    except:  # noqa: E722
        # If the construction of the new dataset fails, the created
        # partition is deleted, to guarantee the integrity of the
        # collection.
        _rm(fs, dirname)
        raise


def _load_and_apply_indexer(
    args: tuple[tuple[tuple[str, int], ...], list[slice]],
    *,
    delayed: bool,
    fs: fsspec.AbstractFileSystem,
    partition_handler: partitioning.Partitioning,
    partition_properties: PartitioningProperties,
    selected_variables: Iterable[str] | None,
) -> list[dataset.Dataset]:
    """Load a partition and apply its indexer.

    Args:
        args: Tuple containing the partition's keys and its indexer.
        fs: The file system that the partition is stored on.
        partition_handler: The partitioning handler.
        partition_properties: The partitioning properties.
        selected_variable: The selected variables to load.

    Returns:
        The list of loaded datasets.
    """
    items: list[slice]
    partition_scheme: tuple[tuple[str, int], ...]

    partition_scheme, items = args
    partition: str = join_path(
        partition_properties.dir,
        partition_handler.join(partition_scheme, fs.sep))
    zds: dataset.Dataset = open_zarr_group(
        partition, fs, delayed=delayed, selected_variables=selected_variables)
    return list(
        zds.isel({partition_properties.dim: indexer}) for indexer in items)
